/*
 *  Copyright (c) 2008 Luca Abeni
 *
 *  This is free software; see GPL.txt
 */
#include <stdarg.h>
#include <unistd.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "avstreamer.h"
#include "rtp.h"
#include "in.h"
#include "sdp.h"
#include "dbg.h"
#include "file_feed.h"

#define THRESHOLD 1

struct insession_descr {
    struct sdp_descr sdp;
    const char *url;
    struct rtp_context *input;
    struct insession *in;

    struct rtp_session *out[MAX_MEDIA];
};

static void send_dyn_buff(struct rtp_session *rtp, uint8_t *buff, int len)
{   
    uint8_t *p = buff;
    int l;

    if (rtp == NULL) {
        return;
    }

    while (p + 4 <= buff + len) {
        l = (p[0] << 24) | (p[1] << 16) | (p[2] << 8) | (p[3]);
        p += 4;
        if ((p[1] >= 200) && (p[1] <= 204)) {
            rtcp_send(rtp, p, l);
        } else {
            rtp_send(rtp, p, l);
        }
        p += l;
    }
}

static int send_handler(int stream, void *c)
{
    struct insession_descr *s = c;
    uint8_t *buff; 
    int len, frames;
    
    len = rtp_get_buffer(s->input, stream, &buff);
    if (buff) {
        send_dyn_buff(s->out[stream], buff, len);
    }
    rtp_release_buffer(s->input, stream, buff);
    
    frames = rtp_dequeue_frame(s->input, stream);
    dbg_level_inc();
    dbg_print("Queued frames decreased to %d\n", frames);
    if (frames < THRESHOLD) {
        int res;
 
        dbg_print("Now have to read again!!!\n");
        res = send_frame(s->input, stream);
        if (res == -1) {
            res = rtp_stream_destroy(s->input);
            if (res > 0) {
                s->input = NULL;
                if (s->sdp.media[0] == NULL) {
                    /* This input session has already been released...
                     * We can destroy it!
                     */
                    free(s->in);
                }
            }
        }
    }
    dbg_level_dec();

    return 0;
}



/*
 * Methods
 */
static struct insession *insession_avs_get(const char *url)
{
    struct insession *s;
    int res;
    const char *full_fname;
    static char buff[1024];

    full_fname = full_file_name(url);
    s = malloc(sizeof(struct insession) + sizeof(struct insession_descr));
    if (s == 0) {
        goto err2;
    }
    memset(s, 0, sizeof(struct insession) + sizeof(struct insession_descr));
    s->ops = &IS_AVS;
    s->ctx = (struct insession_descr *)((uint8_t *)s + sizeof(struct insession));

    s->ctx->input = rtp_stream_create(full_fname, NULL, NULL, 50 * 1000);	/* FIXME: maybe should be larger for TCP? */
    free((void *)full_fname);
    if (s->ctx->input == NULL) {
        goto err1;
    }
    s->ctx->url = strdup(url);
    s->ctx->in = s;
    res = rtp_get_sdp(s->ctx->input, buff, sizeof(buff));
    if (res < 0) {
        goto err0;
    }
    res = sdp_set(&s->ctx->sdp, buff);
    if (res < 0) {
        goto err0;
    }

    return s;

err0:
    rtp_stream_destroy(s->ctx->input);
err1:
    free(s);
err2:
    return NULL;
}

static void release(struct insession_descr *s)
{
    int res;

    sdp_free(&s->sdp);
    res = rtp_stream_destroy(s->input);
    free((void *)s->url);
    if (res != 0) {
        s->input = NULL;
        free(s->in);
    }
}

static struct insession *insession_avs_allocate(const char *url)
{
    return NULL;
}

static const char *get_url(struct insession_descr *s)
{
    return s->url;
}

static int get_sdp(char *buff, int size, struct insession_descr *s)
{
    return sdp_get(buff, size, &s->sdp);
}

static struct media *media(struct insession_descr *s, int i)
{
    if ((i > MAX_MEDIA) || (s == NULL)) {
        return NULL;
    }

    return s->sdp.media[i];
}

static void activate(struct insession_descr *s, void *h)
{
    rtp_set_completion(s->input, send_handler, s, h);
}

static int set_sdp(struct insession_descr *s, const char *sdp)
{
    fprintf(stderr, "InSession Set SDP: This should never be called!!!\n");

    return -1;
}

static void play(struct insession_descr *s, struct rtp_session *r, int i)
{
    r->next = s->out[i];
    s->out[i] = r;
    send_frame(s->input, i);
}

static void stop(struct insession_descr *s, struct rtp_session *r)
{
    int i;

    for (i = 0; i < MAX_MEDIA; i++) {
        if (s->out[i] == r) {
            s->out[i] = r->next;
            return;
        } else {
            struct rtp_session *p;

            p = s->out[i];
            while(p != NULL) {
                if (p->next == r) {
                    p->next = r->next;
                    return;
                } else {
                    p = p->next;
                }
            }
        }
    }
}

struct insession_methods IS_AVS = {
    .get = insession_avs_get,
    .allocate = insession_avs_allocate,
    .release = release,
    .activate = activate,
    .get_url = get_url,
    .get_sdp = get_sdp,
    .set_sdp = set_sdp,
    .media = media,
    .play = play,
    .stop = stop,
};
