/*
 *  Copyright (c) 2008 Luca Abeni
 *
 *  This is free software; see GPL.txt
 */
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if.h>     /* For struct ifreq */
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "net.h"
#include "dbg.h"

#define MAX_PENDING_REQUESTS 2

int fd_addr(int fd)
{
    struct sockaddr_in addr;
    int size, res;

    size = sizeof(struct sockaddr_in);
    res = getsockname(fd, (struct sockaddr *)&addr, &size);
    if (res < 0) {
        return 0;
    }

    return addr.sin_addr.s_addr;
}

int udp_port(struct udp_connection *r)
{
    struct sockaddr_in addr;
    int size, res;

    size = sizeof(struct sockaddr_in);
    res = getsockname(r->s, (struct sockaddr *)&addr, &size);
    if (res < 0) {
        return 0;
    }

    return ntohs(addr.sin_port);
}

const char *iface_addr(const char *iface)
{
    int s, res;
    struct ifreq iface_request;
    struct sockaddr_in *sin;
    char buff[512];

    s = socket(AF_INET, SOCK_DGRAM, 0);
    if (s < 0) {
        return NULL;
    }

    memset(&iface_request, 0, sizeof(struct ifreq));
    sin = (struct sockaddr_in *)&iface_request.ifr_addr;
    strcpy(iface_request.ifr_name, iface);
    /* sin->sin_family = AF_INET); */
    res = ioctl(s, SIOCGIFADDR, &iface_request);
    if (res < 0) {
        perror("ioctl(SIOCGIFADDR)");
        close(s);

        return NULL;
    }
    close(s);

    inet_ntop(AF_INET, &sin->sin_addr, buff, sizeof(buff));

    return strdup(buff);
}



static int udp_bind(int s, const char *addr, int port)
{
    struct sockaddr_in my_addr;
    int res;

    /* FIXME: If addr is multicast, join the group! */
    memset(&my_addr, 0, sizeof(struct sockaddr_in));
    my_addr.sin_family = AF_INET;
    my_addr.sin_port = htons(port);
    res = inet_aton(addr, &my_addr.sin_addr);
    if (res == 0) {
        return -1;
    }

    return bind(s, (struct sockaddr *)&my_addr, sizeof(struct sockaddr_in));
}

int server_init(const char *iface, int port)
{
    struct sockaddr_in server_addr;
    int fd, res;

    fd = socket(AF_INET, SOCK_STREAM, 0);
    if (fd < 0) {
        perror ("Socket");

        return -1;
    }

#if 0
    val = 1;
    setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &tmp, sizeof(int));
#endif
    memset(&server_addr, 0, sizeof(struct sockaddr_in));
    server_addr.sin_family = AF_INET;
    server_addr.sin_port = htons(port);
    res = inet_aton(iface, &server_addr.sin_addr);
    if (res == 0) {
        perror("Inet AToN");
        server_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    }

    dbg_print("Connecting server on port %d %d:\t", port, htons(port));

    res = bind(fd, (struct sockaddr *)&server_addr, sizeof(struct sockaddr_in));
    if (res < 0) {
        fprintf(stderr,"Port %d:\t", port);
        perror("Bind");
        close(fd);

        return -1;
    }

    if (listen(fd, MAX_PENDING_REQUESTS) < 0) {
        perror ("listen");
        close(fd);

        return -1;
    }
#if 0
    res = fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) | O_NONBLOCK);
    if (res < 0) {
        perror("FCntl");
    }
#endif

    return fd;
}

int new_connection(int s, int *client_ip)
{
    struct sockaddr_in from_addr;
    int fd;
    socklen_t len;

    len = sizeof(struct sockaddr_in);
    fd = accept(s, (struct sockaddr *)&from_addr, &len);
    if (fd < 0) {
        perror("Accept");

        return -1;
    }
    if (len != sizeof(struct sockaddr_in)) {
        fprintf(stderr, "Strange kind of SRC address: len = %d\n", len);
    }
    *client_ip = from_addr.sin_addr.s_addr;

    return fd;
}

struct udp_connection *udp_sock(const char *addr, int port, const char *iface, int local_port)
{
    int res;
    struct udp_connection *r;
    struct in_addr in;

    r = malloc(sizeof(struct udp_connection));
    if (r == NULL) {
        return NULL;
    }
    r->s = socket(PF_INET, SOCK_DGRAM, 0);
    if (r->s < 0) {
        free(r);

        return NULL;
    }
    if (iface) {
        udp_bind(r->s, iface, local_port);
    }

    r->port = htons(port);
    res = inet_aton(addr, &in);
    if (res == 0) {
        close(r->s);
        free(r);

        return NULL;
    }
    r->addr = in.s_addr;

    return r;
}

int udp_send(struct udp_connection *r, const char *buff, int len)
{
    struct sockaddr_in dst;
    int res;

    dst.sin_family = AF_INET;
    dst.sin_port = r->port;
    dst.sin_addr.s_addr = r->addr;
    res = sendto(r->s, buff, len, 0, (struct sockaddr *)&dst, sizeof(struct sockaddr_in));

    return res;
}

struct udp_connection *udp_insock(const char *addr, int port)
{
    struct udp_connection *r;
    int res;

    r = malloc(sizeof(struct udp_connection));
    if (r == NULL) {
        return NULL;
    }

    r->s = socket(PF_INET, SOCK_DGRAM, 0);
    if (r->s < 0) {
        return NULL;
    }

    res = udp_bind(r->s, addr, port);
    if (res < 0) {
        fprintf(stderr,"Port %d:\t", port);
        perror("Bind");
        close(r->s);
        free(r);

        return NULL;
    }

    return r;
}

void udp_close(struct udp_connection *c)
{
    close(c->s);
    free(c);
}

const char *net_ascii_addr(unsigned long int addr)
{
    struct in_addr in;

    in.s_addr = addr;
    return strdup(inet_ntoa(in));
}

int connection_fd(struct udp_connection *r)
{
    return r->s;
}
