/*
 *  Copyright (c) 2008 Luca Abeni
 *
 *  This is free software; see GPL.txt
 */
#include <unistd.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "engine.h"
#include "net.h"
#include "rtp.h"
#include "rtcp.h"
#include "stats.h"

static int private_rtp_port = 9752;

void write_rtp_over_rtsp_header(int fd, int id, int len)
{
    uint8_t h[4];

    h[0] = '$';
    h[1] = id;
    h[2] = len >> 8;
    h[3] = len;
    write(fd, h, 4);
}

static int rtcp_handler(int fd, void *h)
{
    struct rtp_session *r;
    static uint8_t buff[4096];
    int res;

    res = read(fd, buff, 4096);
    if (res < 0) {
        perror("Read");

        return -1;
    }
    if (res > 1500) {
        fprintf(stderr, "Uhmmm... %d\n", res);
    }
    if (res == 4096) {
        return -1;
    }
    r = h;
    if (r->stats) {
        rtcp_parse(buff, r->stats);
    }

    return 0;
}

static int rtcp_alarm_handler(uint64_t t, void *c)
{
    struct rtp_session *r;

    r = c;
    if (r->stats) {
        print_rtcp_stats(t, r->stats);
        r->alarm_h = timed_event_new(r->h, t + 3000000, rtcp_alarm_handler, c);

        return 0;
    }

    return -1;
}

struct rtp_session *rtp_session_create(void *h, const char *dst, int id, int tcp_fd, const char *iface)
{
    struct rtp_session *r;

    r = malloc(sizeof(struct rtp_session));
    if (r == NULL) {
        return NULL;
    }
    memset(r, 0, sizeof(struct rtp_session));
    if (tcp_fd != 0) {
        r->stream_index = id;
        r->tcp_fd = tcp_fd;
    } else {
        r->sock = udp_sock(dst, id, iface, private_rtp_port);
        if (r->sock == NULL) {
            free(r);
            return NULL;
        }
        r->ctrl_sock = udp_sock(dst, id + 1, iface, private_rtp_port + 1);
        if (r->ctrl_sock == NULL) {
            udp_close(r->sock);
            free(r);
            return NULL;
        }
        private_rtp_port = udp_port(r->sock) + 2;
        event_new(h, connection_fd(r->ctrl_sock), rtcp_handler, r);
    }
    r->h = h;
    r->stats = rtcp_new();
    r->alarm_h = timed_event_new(h, -3000000, rtcp_alarm_handler, r);
    init_rtcp_stats(r->stats, dst, id);
    r->next = NULL;

    return r;
}

void rtp_session_close(struct rtp_session *r)
{
    if (r == NULL) {
        return;
    }
    timed_event_del(r->h, r->alarm_h);
    if (r->tcp_fd == 0) {
        event_del(r->h, connection_fd(r->ctrl_sock));
        udp_close(r->sock);
        udp_close(r->ctrl_sock);
    }
    cleanup_rtcp_stats(r->stats);
    free(r);
}

struct rtp_session *rtp_insession_create(void *h, const char *dst, int port)
{
    struct rtp_session *r;

    r = malloc(sizeof(struct rtp_session));
    if (r == NULL) {
        return NULL;
    }
    memset(r, 0, sizeof(struct rtp_session));

    r->sock = udp_insock(dst, port);
    if (r->sock == NULL) {
        free(r);
        return NULL;
    }
    r->ctrl_sock = udp_insock(dst, port + 1);
    if (r->ctrl_sock == NULL) {
        udp_close(r->sock);
        free(r);
        return NULL;
    }
    r->h = h;

    return r;
}

int rtp_insession_handle(struct rtp_session *s, int (*h1)(int s, void *c), int (*h2)(int s, void *c), void *c)
{
    int res;

    res = event_new(s->h, connection_fd(s->sock), h1, c);
    if (res < 0) {
        return res;
    }

    return event_new(s->h, connection_fd(s->ctrl_sock), h2, c);
}

/* FIXME: Find some way to merge in rtp_session_close()... */
void rtp_insession_release(struct rtp_session *s)
{
    if (s == NULL) {
        return;
    }
    event_del(s->h, connection_fd(s->sock));
    event_del(s->h, connection_fd(s->ctrl_sock));
    rtp_session_close(s);
}

int rtp_port(struct rtp_session *r)
{
#if 0
    int res;
    res = udp_port(r->ctrl_sock) - 1;
    fprintf(stderr, "Returning RTP %d\n", res);

    return res;
#else
    return udp_port(r->sock);
#endif
}

int rtcp_port(struct rtp_session *r)
{
    int res;
    res = udp_port(r->ctrl_sock);
    /*fprintf(stderr, "Returning RTCP %d\n", res);*/

    return res;
}

int rtp_send(struct rtp_session *r, uint8_t *b, int size)
{
    if (r->stats) {
        r->stats->sum += size;
        r->stats->sent_pkts++;
    }
    if (r->tcp_fd) {
        write_rtp_over_rtsp_header(r->tcp_fd, r->stream_index, size);
	return write(r->tcp_fd, b, size);
    }
    return udp_send(r->sock, (char *)b, size);
}

int rtcp_send(struct rtp_session *r, uint8_t *b, int size)
{
    if (r->tcp_fd) {
        write_rtp_over_rtsp_header(r->tcp_fd, r->stream_index + 1, size);
	return write(r->tcp_fd, b, size);
    }
    return udp_send(r->ctrl_sock, (char *)b, size);
}
