/*
 *  Copyright (c) 2008 Luca Abeni
 *
 *  This is free software; see GPL.txt
 */
#include <time.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>

#include "net.h"
#include "rtp.h"
#include "in.h"
#include "session.h"
#include "file_feed.h"
#include "rtsp_commands.h"
#include "dbg.h"

static struct {
    const char *cmd_name;
    int cmd_id;
    void (*cmd_action)(char *buff, int buff_size, struct rtsp_request *r);
} cmds_array[] = {
    {
        .cmd_name = "OPTIONS",
        .cmd_id = RTSP_CMD_OPTIONS,
        .cmd_action = rtsp_cmd_options,
    },
    {
        .cmd_name = "DESCRIBE",
        .cmd_id = RTSP_CMD_DESCRIBE,
        .cmd_action = rtsp_cmd_describe,
    },
    {
        .cmd_name = "SETUP",
        .cmd_id = RTSP_CMD_SETUP,
        .cmd_action = rtsp_cmd_setup,
    },
    {
        .cmd_name = "PLAY",
        .cmd_id = RTSP_CMD_PLAY,
        .cmd_action = rtsp_cmd_play,
    },
    {
        .cmd_name = "TEARDOWN",
        .cmd_id = RTSP_CMD_TEARDOWN,
        .cmd_action = rtsp_cmd_teardown,
    },
    {
        .cmd_name = NULL,
        .cmd_id = -1,
        .cmd_action = NULL,
    }
};

static const char *rtsp_mode(int m)
{
    switch (m) {
        case RTP_UDP:
            return "RTP/AVP";
        case RTP_TCP:
            return "RTP/AVP/TCP";
        case RAW_UDP:
            return "RAW/RAW/UDP";
        case TS_UDP:
            return "MP2T/H2221/UDP";
        default:
            return "Unknown";
    }
}

static const char *transport_string(struct rtp_session *r, struct rtsp_request *req)
{
    static char res[512];
    
    switch (req->streaming_mode) {
        case RTP_UDP:
            snprintf(res, 512, "server_port=%d-%d;client_port=%d-%d",
                     rtp_port(r), rtcp_port(r), req->rtp_port, req->rtcp_port);
            break;
        case RTP_TCP:
            snprintf(res, 512, "interleaved=%d-%d",
                     req->rtp_channel_id, req->rtcp_channel_id);
            break;
        case RAW_UDP:
        case TS_UDP:
            snprintf(res, 512, "server_port=%d;client_port=%d",
                     rtp_port(r), req->rtp_port);
            break;
        default:
            return "Unknown";
    }

    return res;
}
static char const *rtsp_date(void)
{
    static char buff[200];
    time_t tt = time(NULL);

    strftime(buff, 200, "Date: %a, %b %d %Y %H:%M:%S GMT\r\n", gmtime(&tt));

    return buff;
}

#if 0
/* FIXME: Rewrite using the commands array... */
static char const* rtsp_commands =
                   "OPTIONS, DESCRIBE, SETUP, TEARDOWN, PLAY, PAUSE";
#else
char *rtsp_commands(void)
{
    static char *res;
    int i;

    if (res != NULL) {
        return res;
    }
    res = malloc(512);
    if (res != NULL) {
        res[0] = 0;
	i = 0;
        while (cmds_array[i].cmd_name != NULL) {
            strcat(res, cmds_array[i].cmd_name);
            strcat(res, " ");
            i++;
        }
    }

    return res;
}
#endif
int rtsp_command(const char *cmd)
{
    int i;

    i = 0;
    while (cmds_array[i].cmd_name != NULL) {
        if (strcmp(cmds_array[i].cmd_name, cmd) == 0) {
            return cmds_array[i].cmd_id;
        }
	i++;
    }

    return -1;
}

void rtsp_cmd_unknown(char *buff, int buff_size, struct rtsp_request *r)
{
    int cseq = r->cseq;
    int res;

    res = snprintf(buff, buff_size,
                   "RTSP/1.0 405 Method Not Allowed\r\nCSeq: %d\r\n%sAllow: %s\r\n\r\n",
                   cseq, rtsp_date(), rtsp_commands());
    buff[res] = 0;
}
void rtsp_cmd_error(char *buff, int buff_size)
{
    int res;

    res = snprintf(buff, buff_size,
                   "RTSP/1.0 400 Bad Request\r\n%sAllow: %s\r\n\r\n",
                   rtsp_date(), rtsp_commands());
    buff[res] = 0;
}

void rtsp_cmd_options(char *buff, int buff_size, struct rtsp_request *r)
{
    int res;
    int cseq = r->cseq;

    res = snprintf(buff, buff_size,
                   "RTSP/1.0 200 OK\r\nCSeq: %d\r\n%sPublic: %s\r\n\r\n",
                   cseq, rtsp_date(), rtsp_commands());
    buff[res] = 0;
}

static struct insession *search_from_url(const char *url)
{
    struct insession *s;
    int res;
    const char *sdp;

    s = insession_get(url);
    if (s != NULL) {
        return s;
    }

    sdp = sdp_from_file(url);
    if (sdp == NULL) {
        return NULL;
    }
    s = insession_allocate(url);
    if (s == NULL) {
        free((char *)sdp);
        return NULL;
    }
    res = insession_set_sdp(s, sdp);
    free((char *)sdp);
    if (res < 0) {
        insession_release(s);
        return NULL;
    }

    return s;
}

#define SDP_SIZE 2048
void rtsp_cmd_describe(char *buff, int buff_size, struct rtsp_request *r)
{
    int res;
    void *s;
    const char *rtsp_url;
    static char sdp[SDP_SIZE];

    dbg_print("Describe %s %s\n", r->stream_name, r->track_id);
    /* FIXME: Implement authentication */
    s = search_from_url(r->stream_name);
    if (s == NULL) {
        res = snprintf(buff, buff_size,
                       "RTSP/1.0 404 Stream Not Found\r\n"
                       "CSeq: %d\r\n%s\r\n",
                       r->cseq, rtsp_date());
        buff[res] = 0;

	return;
    }
    rtsp_url = insession_get_url(s);
    res = insession_get_sdp(sdp, SDP_SIZE, s);
    if (res < 0) {
        res = snprintf(buff, buff_size,
                       "RTSP/1.0 404 Stream Not Found\r\n"
                       "CSeq: %d\r\n%s\r\n",
                       r->cseq, rtsp_date());
        buff[res] = 0;
    } else {
        res = snprintf(buff, buff_size,
                       "RTSP/1.0 200 OK\r\nCSeq: %d\r\n%s"
                       /*"Content-Base: %s/\r\n"*/
                       "Content-Type: application/sdp\r\n"
                       "Content-Length: %d\r\n\r\n%s",
                       r->cseq, rtsp_date(), /*rtsp_url,*/ (int)strlen(sdp), sdp);
    }
    buff[res] = 0;
    insession_release(s);
}

void rtsp_cmd_setup(char *buff, int buff_size, struct rtsp_request *r)
{
    struct insession *s;
    int res;
    unsigned int track_id;
    struct rtp_session *out;
    const char *iface;

    res = sscanf(r->track_id, "TrackID=%d", &track_id);
    if (res < 1) {
        res = sscanf(r->track_id, "streamid=%d", &track_id);
        track_id++;
    }
    dbg_print("Setup %s %s - %u\n", r->stream_name, r->track_id, track_id);
    dbg_print("RT(C)P Ports: %hu %hu\n", r->rtp_port, r->rtcp_port);
    dbg_print("RT(C)P Channel ids: %u %u\n", r->rtp_channel_id, r->rtcp_channel_id);
    dbg_print("Streaming mode: %d\n", r->streaming_mode);
    dbg_print("Destination: %s\n", r->dst);
    dbg_print("From: %x\n", r->from);
    if (r->session_id == 0) {
      s = search_from_url(r->stream_name);
      r->session_id = session_new(s);
    } else {
      s = insession_from_id(r->session_id);
    }
    if ((s == NULL) || (insession_media(s, track_id - 1) == NULL)) {
        res = snprintf(buff, buff_size,
                       "RTSP/1.0 404 Stream Not Found\r\n"
                       "CSeq: %d\r\n%s\r\n",
                       r->cseq, rtsp_date());
        buff[res] = 0;

	return;
    }

    if ((r->streaming_mode == RTP_TCP) && (r->rtp_channel_id == 0)) {
        static int default_channel_id;

        dbg_print("Crap! A dumb client requested RTP over TCP without interleaving!\n");
        dbg_print("Setting a default channel id for the idiot...\n");
        r->rtp_channel_id = default_channel_id;
        r->rtcp_channel_id = r->rtp_channel_id + 1;
        default_channel_id += 2;
    }
    if (r->dst == NULL) {
        r->dst = net_ascii_addr(r->from);
    }

    insession_activate(s, r->handle);

    iface = net_ascii_addr(r->local_addr);
    out = rtp_session_create(r->handle, r->dst, r->streaming_mode == RTP_UDP ? r->rtp_port : r->rtp_channel_id,
                             r->streaming_mode == RTP_UDP ? 0 : r->fd, iface);
    free(iface);
    if (out == NULL) {
        insession_release(s);
        res = snprintf(buff, buff_size,
                       "RTSP/1.0 500 Internal Server Error\r\n"
                       "CSeq: %d\r\n%s\r\n",
                       r->cseq, rtsp_date());
        buff[res] = 0;

	return;
    }

    session_add_stream(r->session_id, out, track_id - 1); /*connection->rtp[tid - 1] = r->rtp;*/
    
    res = snprintf(buff, buff_size,
                   "RTSP/1.0 200 OK\r\nCSeq: %d\r\n"
                   "%s"
                   "Transport: %s;%s;destination=%s;%s\r\n"
                   "Session: %d\r\n\r\n",
                   r->cseq, rtsp_date(), rtsp_mode(r->streaming_mode),
                   r->streaming_mode & MULTICAST ? "multicast" : "unicast",
                   r->dst, transport_string(out, r), r->session_id);
    buff[res] = 0;
}

void rtsp_cmd_play(char *buff, int buff_size, struct rtsp_request *r)
{
    int res;
    struct insession *s;
    
    dbg_print("Play %s\n", r->stream_name);
    dbg_print("RT(C)P Ports: %hu %hu\n", r->rtp_port, r->rtcp_port);
    dbg_print("RT(C)P Channel ids: %u %u\n", r->rtp_channel_id, r->rtcp_channel_id);
    dbg_print("Streaming mode: %d\n", r->streaming_mode);
    dbg_print("Destination: %s\n", r->dst);
    dbg_print("From: %x\n", r->from);

    s = insession_from_id(r->session_id);
    if ((s == NULL) || (strcmp(insession_get_url(s), r->stream_name))) {
        res = snprintf(buff, buff_size,
                       "RTSP/1.0 404 Stream Not Found\r\n"
                       "CSeq: %d\r\n%s\r\n",
                       r->cseq, rtsp_date());
        buff[res] = 0;

	return;
    }
    session_play(r->session_id);

    res = snprintf(buff, buff_size,
                   "RTSP/1.0 200 OK\r\nCSeq: %d\r\n"
                   "%s"
                   "Session: %d\r\n\r\n",
                   r->cseq, rtsp_date(), 
                   r->session_id);
    buff[res] = 0;
}

void rtsp_cmd_teardown(char *buff, int buff_size, struct rtsp_request *r)
{
    int res;
    struct insession *s;
    
    dbg_print("Teardown %s\n", r->stream_name);
    dbg_print("RT(C)P Ports: %hu %hu\n", r->rtp_port, r->rtcp_port);
    dbg_print("RT(C)P Channel ids: %u %u\n", r->rtp_channel_id, r->rtcp_channel_id);
    dbg_print("Streaming mode: %d\n", r->streaming_mode);
    dbg_print("Destination: %s\n", r->dst);
    dbg_print("From: %x\n", r->from);

    s = insession_from_id(r->session_id);
    if ((s == NULL) || (strcmp(insession_get_url(s), r->stream_name))) {
        res = snprintf(buff, buff_size,
                       "RTSP/1.0 404 Stream Not Found\r\n"
                       "CSeq: %d\r\n%s\r\n",
                       r->cseq, rtsp_date());
        buff[res] = 0;

	return;
    }
    session_stop(r->session_id);
    session_delete(r->session_id);

    res = snprintf(buff, buff_size,
                   "RTSP/1.0 200 OK\r\nCSeq: %d\r\n"
                   "%s"
                   "Session: %d\r\n\r\n",
                   r->cseq, rtsp_date(), 
                   r->session_id);
    buff[res] = 0;
}
